--------------------------------------------------------------------------------
-- A new design for multirec using data kinds and kind polymorphism
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE DefaultSignatures      #-}
{-# LANGUAGE TypeOperators          #-}
{-# LANGUAGE RankNTypes             #-}
{-# LANGUAGE MultiParamTypeClasses  #-}

module Library.Multirec where


import qualified Data.Foldable as F

--------------------------------------------------------------------------------
-- Universe
--------------------------------------------------------------------------------

data Un s i = K s | I i | U | TAG (Un s i) i
            -- | C MC (Un s) -- omitting because it adds nothing new
            | SUM  (Un s i) (Un s i)
            | PROD (Un s i) (Un s i)
            -- and this actually works:
            | COMP (s -> s) (Un s i)

--------------------------------------------------------------------------------
-- Interpretation as a GADT
--------------------------------------------------------------------------------

data In (u :: Un * i) (r :: i -> *) (ix :: i) where
  In_K :: a -> In (K a) r i
  In_I :: r j -> In (I j) r i
  In_T :: In a r i -> In (TAG a i) r i
  In_U :: In U r i
  In_P :: In a r i -> In b r i -> In (PROD a b) r i
  In_L :: In a r i -> In (SUM a b) r i
  In_R :: In b r i -> In (SUM a b) r i
  In_C :: f (In a r i) -> In (COMP f a) r i

--------------------------------------------------------------------------------
-- Map
--------------------------------------------------------------------------------

class GMap phi (f :: Un * i) where
  gmap :: (forall ix. phi ix -> r ix -> s ix) -> phi ix -> In f r ix -> In f s ix

instance GMap phi (K a) where
  gmap _ _ (In_K a) = In_K a

instance (El phi ix) => GMap phi (I ix) where
  gmap f _ (In_I r) = In_I (f proof r)

instance (GMap phi a) => GMap phi (TAG a i) where
  gmap f p (In_T a) = In_T (gmap f p a)

instance GMap phi U where
  gmap f _ In_U = In_U

instance (GMap phi a, GMap phi b) => GMap phi (SUM a b) where
  gmap f p (In_L a) = In_L (gmap f p a)
  gmap f p (In_R b) = In_R (gmap f p b)

instance (GMap phi a, GMap phi b) => GMap phi (PROD a b) where
  gmap f p (In_P a b) = In_P (gmap f p a) (gmap f p b)

instance (Functor f, GMap phi a) => GMap phi (COMP f a) where
  gmap f p (In_C fa) = In_C (fmap (gmap f p) fa)


type Algebra phi r = forall ix. phi ix -> In (PF phi) r ix -> r ix

cata :: (Fam phi, GMap phi (PF phi)) => Algebra phi r -> phi ix -> ix -> r ix
cata f p = f p . gmap (\p (I0 x) -> cata f p x) p . from p

--------------------------------------------------------------------------------
-- Size (as a catamorphism)
--------------------------------------------------------------------------------

newtype Int1 r = Int1 { unInt1 :: Int }

class SizeAlg (f :: Un * i) where
  sizeAlg :: In f Int1 ix -> Int

instance SizeAlg U     where sizeAlg _ = 0
instance SizeAlg (K a) where sizeAlg _ = 0
instance SizeAlg (I i) where sizeAlg (In_I (Int1 n)) = n + 1
instance (SizeAlg a) => SizeAlg (TAG a i) where
  sizeAlg (In_T a) = sizeAlg a
instance (SizeAlg a, SizeAlg b) => SizeAlg (PROD a b) where
  sizeAlg (In_P a b) = sizeAlg a + sizeAlg b
instance (SizeAlg a, SizeAlg b) => SizeAlg (SUM a b)  where
  sizeAlg (In_L a) = sizeAlg a
  sizeAlg (In_R b) = sizeAlg b
instance (F.Foldable f, SizeAlg a) => SizeAlg (COMP f a)  where
  sizeAlg (In_C fa) = F.foldr (\a b -> sizeAlg a + b) 0 fa


size :: (Fam phi, GMap phi (PF phi), SizeAlg (PF phi)) => phi ix -> ix -> Int
size p = unInt1 . cata (\_ -> Int1 . sizeAlg) p

--------------------------------------------------------------------------------
-- Conversions to/from user datatypes
--------------------------------------------------------------------------------

newtype I0 r = I0 { unI0 :: r }

class Fam (phi :: * -> *) where
  type PF phi :: Un * *
  from :: phi ix -> ix -> In (PF phi) I0 ix
  to   :: phi ix -> In (PF phi) I0 ix -> ix

class El (phi :: i -> *) (ix :: i) where
  proof :: phi ix

--------------------------------------------------------------------------------
-- Example encoding: Zig/Zag
--------------------------------------------------------------------------------

data Zig = Zig Zag | ZigEnd
data Zag = Zag [Zig]

data ZigZag ix where
  ZigZag_Zig :: ZigZag Zig
  ZigZag_Zag :: ZigZag Zag

instance Fam ZigZag where
  type PF ZigZag = SUM (TAG (SUM (I Zag) U)   Zig)
                       (TAG (COMP [] (I Zig)) Zag)

  from ZigZag_Zig (Zig zag)  = In_L (In_T (In_L (In_I (I0 zag))))
  from ZigZag_Zig (ZigEnd)   = In_L (In_T (In_R In_U))
  from ZigZag_Zag (Zag zigs) = In_R (In_T (In_C (map (In_I . I0) zigs)))

  to ZigZag_Zig (In_L (In_T (In_L (In_I (I0 zag))))) = Zig zag
  to ZigZag_Zig (In_L (In_T (In_R In_U)))            = ZigEnd
  to ZigZag_Zag (In_R (In_T (In_C zigs)))            = Zag (map (\(In_I (I0 x)) -> x) zigs)

instance El ZigZag Zig where
  proof = ZigZag_Zig

instance El ZigZag Zag where
  proof = ZigZag_Zag

zigZag :: Zig
zigZag = Zig (Zag [Zig (Zag [ZigEnd]), ZigEnd])

testSize :: Int
testSize = size ZigZag_Zig zigZag
