--------------------------------------------------------------------------------
-- A new design for regular using data kinds and kind polymorphism
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE DefaultSignatures      #-}
{-# LANGUAGE TypeOperators          #-}

module Library.Regular ( module Library.Regular
                       , module Auxiliary.Metadata ) where


import Auxiliary.Utils
import Auxiliary.Metadata

--------------------------------------------------------------------------------
-- Universe
--------------------------------------------------------------------------------

data Un s = K s | I | U
          | C MetaCons (Un s)
          | SUM (Un s) (Un s) | PROD (Un s) (Un s)

--------------------------------------------------------------------------------
-- Interpretation as a GADT
--------------------------------------------------------------------------------

data In (u :: Un *) (r :: *) where
  In_K :: a -> In (K a) r
  In_I :: r -> In I r
  In_U :: In U r
  -- We could have a `SingI name` constraint below, but I don't see the
  -- advantage of that, compared to putting those constraints in the generic
  -- function definition.
  In_C :: In a r -> In (C mc a) r
  In_P :: In a r -> In b r -> In (PROD a b) r
  In_L :: In a r -> In (SUM a b) r
  In_R :: In b r -> In (SUM a b) r

--------------------------------------------------------------------------------
-- Map
--------------------------------------------------------------------------------

class GMap (f :: Un *) where
  gmap :: (r -> s) -> In f r -> In f s

instance GMap (K a) where
  gmap _ (In_K a) = In_K a

instance GMap I where
  gmap f (In_I r) = In_I (f r)

instance GMap U where
  gmap f In_U = In_U

instance (GMap a) => GMap (C c a) where
  gmap f (In_C x) = In_C (gmap f x)

instance (GMap a, GMap b) => GMap (SUM a b) where
  gmap f (In_L a) = In_L (gmap f a)
  gmap f (In_R b) = In_R (gmap f b)

instance (GMap a, GMap b) => GMap (PROD a b) where
  gmap f (In_P a b) = In_P (gmap f a) (gmap f b)

cata :: (Regular a, GMap (PF a)) => (In (PF a) b -> b) -> a -> b
cata f = f . gmap (cata f) . from

--------------------------------------------------------------------------------
-- Size (as a catamorphism)
--------------------------------------------------------------------------------

class SizeAlg (f :: Un *) where
  sizeAlg :: In f Int -> Int

instance SizeAlg U      where sizeAlg _ = 0
instance SizeAlg (K a)  where sizeAlg _ = 0
instance SizeAlg I      where sizeAlg (In_I i) = 1 + i
instance (SizeAlg a) => SizeAlg (C c a) where
  sizeAlg (In_C a) = sizeAlg a
instance (SizeAlg a, SizeAlg b) => SizeAlg (PROD a b) where
  sizeAlg (In_P a b) = sizeAlg a + sizeAlg b
instance (SizeAlg a, SizeAlg b) => SizeAlg (SUM a b)  where
  sizeAlg (In_L a) = sizeAlg a
  sizeAlg (In_R b) = sizeAlg b

size :: (Regular a, GMap (PF a), SizeAlg (PF a)) => a -> Int
size = cata sizeAlg

--------------------------------------------------------------------------------
-- Show
--------------------------------------------------------------------------------

class GShow (f :: Un *) where
  gshow :: (r -> String) -> In f r -> String

instance (Show a) => GShow (K a) where
  gshow _ (In_K a) = show a

instance GShow I where
  gshow f (In_I r) = f r

instance GShow U where
  gshow _ In_U = ""

-- Advantage: we can now match meta-data properties at the type level
instance (SingI name, GShow a) => GShow (C (MC name Prefix isRec) a) where
  gshow f (In_C x) = "(" ++ fromSing (sing :: Sing name) ++ " " ++ gshow f x ++ ")"

-- Note how we assume that the structure under an Infix MC must be a product.
-- This is not encoded in the universe, regrettably, and might lead to "missing
-- instance" errors if we're not careful.
instance (SingI name, GShow a, GShow b)
        => GShow (C (MC name (Infix assoc fix) isRec) (PROD a b)) where
  gshow f (In_C (In_P a b)) = "(" ++ gshow f a ++ " "
                                  ++ fromSing (sing :: Sing name) ++ " "
                                  ++ gshow f b ++ ")"

instance (GShow a, GShow b) => GShow (SUM a b) where
  gshow f (In_L a) = gshow f a
  gshow f (In_R b) = gshow f b

instance (GShow a, GShow b) => GShow (PROD a b) where
  gshow f (In_P a b) = gshow f a ++ " " ++ gshow f b

-- User-facing class
class Show' (a :: *) where
  show' :: a -> String

  default show' :: (Regular a, GShow (PF a)) => a -> String
  show' = gshow show' . from

--------------------------------------------------------------------------------
-- Conversions to/from user datatypes
--------------------------------------------------------------------------------

class Regular (a :: *) where
  type PF a :: Un *

  from :: a -> In (PF a) a
  to   :: In (PF a) a -> a

--------------------------------------------------------------------------------
-- Example encoding: lists
--------------------------------------------------------------------------------
{-
instance Regular [a] where
  type PF [a] = SUM (C (MC "[]" Prefix False) U)
                    (C (MC ":"  (Infix RightAssociative 5) False)
                       (PROD (K a) I))
  
  from []    = In_L (In_C In_U)
  from (h:t) = In_R (In_C (In_P (In_K h) (In_I t)))

  to (In_L (In_C In_U))                     = []
  to (In_R (In_C (In_P (In_K h) (In_I t)))) = h:t

instance Constructor "[]" where
  conName _ = "[]"

instance Constructor ":" where
  conName _ = "(:)"
-}
-- testListSize :: (Int,Int)
-- testListSize = (size "", size "abc")

-- Note the choice between requiring a user-facing class or not...
-- instance (Show a) => Show' [a]

-- testListShow :: String
-- testListShow = show' [1,2,3::Int]
